@@ -3,7 +3,7 @@
  *  Implementation of sequences of expression pairs. */
 
 /*
- *  GiNaC Copyright (C) 1999 Johannes Gutenberg University Mainz, Germany
+ *  GiNaC Copyright (C) 1999-2000 Johannes Gutenberg University Mainz, Germany
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -23,18 +23,24 @@
 #include <algorithm>
 #include <string>
 #include <stdexcept>
+#include <cmath>
 
 #include "expairseq.h"
 #include "lst.h"
+#include "archive.h"
 #include "debugmsg.h"
 #include "utils.h"
 
+#ifndef NO_NAMESPACE_GINAC
 namespace GiNaC {
+#endif // ndef NO_NAMESPACE_GINAC
 
 #ifdef EXPAIRSEQ_USE_HASHTAB
 #error "FIXME: expair_needs_further_processing not yet implemented for hashtabs, sorry. A.F."
 #endif // def EXPAIRSEQ_USE_HASHTAB
 
+GINAC_IMPLEMENT_REGISTERED_CLASS(expairseq, basic)
+
 //////////
 // helper classes
 //////////
@@ -42,7 +48,7 @@ namespace GiNaC {
 class epp_is_less
 {
 public:
-    bool operator()(epp const & lh, epp const & rh) const
+    bool operator()(const epp & lh, const epp & rh) const
     {
         return (*lh).is_less(*rh);
     }
@@ -54,13 +60,13 @@ public:
 
 // public
 
-expairseq::expairseq(expairseq const & other)
+expairseq::expairseq(const expairseq & other)
 {
     debugmsg("expairseq copy constructor",LOGLEVEL_CONSTRUCT);
     copy(other);
 }
 
-expairseq const & expairseq::operator=(expairseq const & other)
+const expairseq & expairseq::operator=(const expairseq & other)
 {
     debugmsg("expairseq operator=",LOGLEVEL_ASSIGNMENT);
     if (this != &other) {
@@ -72,9 +78,9 @@ expairseq const & expairseq::operator=(expairseq const & other)
 
 // protected
 
-void expairseq::copy(expairseq const & other)
+void expairseq::copy(const expairseq & other)
 {
-    basic::copy(other);
+    inherited::copy(other);
     seq=other.seq;
     overall_coeff=other.overall_coeff;
 #ifdef EXPAIRSEQ_USE_HASHTAB
@@ -101,14 +107,14 @@ void expairseq::copy(expairseq const & other)
 // other constructors
 //////////
 
-expairseq::expairseq(ex const & lh, ex const & rh) : basic(TINFO_expairseq)
+expairseq::expairseq(const ex & lh, const ex & rh) : inherited(TINFO_expairseq)
 {
     debugmsg("expairseq constructor from ex,ex",LOGLEVEL_CONSTRUCT);
     construct_from_2_ex(lh,rh);
     GINAC_ASSERT(is_canonical());
 }
 
-expairseq::expairseq(exvector const & v) : basic(TINFO_expairseq)
+expairseq::expairseq(const exvector & v) : inherited(TINFO_expairseq)
 {
     debugmsg("expairseq constructor from exvector",LOGLEVEL_CONSTRUCT);
     construct_from_exvector(v);
@@ -116,8 +122,8 @@ expairseq::expairseq(exvector const & v) : basic(TINFO_expairseq)
 }
 
 /*
-expairseq::expairseq(epvector const & v, bool do_not_canonicalize) :
-    basic(TINFO_expairseq)
+expairseq::expairseq(const epvector & v, bool do_not_canonicalize) :
+    inherited(TINFO_expairseq)
 {
     debugmsg("expairseq constructor from epvector",LOGLEVEL_CONSTRUCT);
     if (do_not_canonicalize) {
@@ -132,16 +138,16 @@ expairseq::expairseq(epvector const & v, bool do_not_canonicalize) :
 }
 */
 
-expairseq::expairseq(epvector const & v, ex const & oc) :
-    basic(TINFO_expairseq), overall_coeff(oc)
+expairseq::expairseq(const epvector & v, const ex & oc) :
+    inherited(TINFO_expairseq), overall_coeff(oc)
 {
     debugmsg("expairseq constructor from epvector,ex",LOGLEVEL_CONSTRUCT);
     construct_from_epvector(v);
     GINAC_ASSERT(is_canonical());
 }
 
-expairseq::expairseq(epvector * vp, ex const & oc) :
-    basic(TINFO_expairseq), overall_coeff(oc)
+expairseq::expairseq(epvector * vp, const ex & oc) :
+    inherited(TINFO_expairseq), overall_coeff(oc)
 {
     debugmsg("expairseq constructor from epvector *,ex",LOGLEVEL_CONSTRUCT);
     GINAC_ASSERT(vp!=0);
@@ -150,6 +156,47 @@ expairseq::expairseq(epvector * vp, ex const & oc) :
     GINAC_ASSERT(is_canonical());
 }
 
+//////////
+// archiving
+//////////
+
+/** Construct object from archive_node. */
+expairseq::expairseq(const archive_node &n, const lst &sym_lst) : inherited(n, sym_lst)
+#ifdef EXPAIRSEQ_USE_HASHTAB
+    , hashtabsize(0)
+#endif
+{
+    debugmsg("expairseq constructor from archive_node", LOGLEVEL_CONSTRUCT);
+    for (unsigned int i=0; true; i++) {
+        ex rest;
+        ex coeff;
+        if (n.find_ex("rest", rest, sym_lst, i) && n.find_ex("coeff", coeff, sym_lst, i))
+            seq.push_back(expair(rest, coeff));
+        else
+            break;
+    }
+    n.find_ex("overall_coeff", overall_coeff, sym_lst);
+}
+
+/** Unarchive the object. */
+ex expairseq::unarchive(const archive_node &n, const lst &sym_lst)
+{
+    return (new expairseq(n, sym_lst))->setflag(status_flags::dynallocated);
+}
+
+/** Archive the object. */
+void expairseq::archive(archive_node &n) const
+{
+    inherited::archive(n);
+    epvector::const_iterator i = seq.begin(), iend = seq.end();
+    while (i != iend) {
+        n.add_ex("rest", i->rest);
+        n.add_ex("coeff", i->coeff);
+        i++;
+    }
+    n.add_ex("overall_coeff", overall_coeff);
+}
+
 //////////
 // functions overriding virtual functions from bases classes
 //////////
@@ -162,12 +209,110 @@ basic * expairseq::duplicate() const
     return new expairseq(*this);
 }
 
-bool expairseq::info(unsigned inf) const
+void expairseq::print(ostream & os, unsigned upper_precedence) const
 {
-    return basic::info(inf);
+    debugmsg("expairseq print",LOGLEVEL_PRINT);
+    os << "[[";
+    printseq(os,',',precedence,upper_precedence);
+    os << "]]";
 }
 
-int expairseq::nops() const
+void expairseq::printraw(ostream & os) const
+{
+    debugmsg("expairseq printraw",LOGLEVEL_PRINT);
+
+    os << "expairseq(";
+    for (epvector::const_iterator cit=seq.begin(); cit!=seq.end(); ++cit) {
+        os << "(";
+        (*cit).rest.printraw(os);
+        os << ",";
+        (*cit).coeff.printraw(os);
+        os << "),";
+    }
+    os << ")";
+}
+
+void expairseq::printtree(ostream & os, unsigned indent) const
+{
+    debugmsg("expairseq printtree",LOGLEVEL_PRINT);
+
+    os << string(indent,' ') << "type=" << typeid(*this).name()
+       << ", hash=" << hashvalue << " (0x" << hex << hashvalue << dec << ")"
+       << ", flags=" << flags
+       << ", nops=" << nops() << endl;
+    for (unsigned i=0; i<seq.size(); ++i) {
+        seq[i].rest.printtree(os,indent+delta_indent);
+        seq[i].coeff.printtree(os,indent+delta_indent);
+        if (i!=seq.size()-1) {
+            os << string(indent+delta_indent,' ') << "-----" << endl;
+        }
+    }
+    if (!overall_coeff.is_equal(default_overall_coeff())) {
+        os << string(indent+delta_indent,' ') << "-----" << endl;
+        os << string(indent+delta_indent,' ') << "overall_coeff" << endl;
+        overall_coeff.printtree(os,indent+delta_indent);
+    }
+    os << string(indent+delta_indent,' ') << "=====" << endl;
+#ifdef EXPAIRSEQ_USE_HASHTAB
+    os << string(indent+delta_indent,' ')
+       << "hashtab size " << hashtabsize << endl;
+    if (hashtabsize==0) return;
+#define MAXCOUNT 5
+    unsigned count[MAXCOUNT+1];
+    for (int i=0; i<MAXCOUNT+1; ++i) count[i]=0;
+    unsigned this_bin_fill;
+    unsigned cum_fill_sq=0;
+    unsigned cum_fill=0;
+    for (unsigned i=0; i<hashtabsize; ++i) {
+        this_bin_fill=0;
+        if (hashtab[i].size()>0) {
+            os << string(indent+delta_indent,' ') 
+               << "bin " << i << " with entries ";
+            for (epplist::const_iterator it=hashtab[i].begin();
+                 it!=hashtab[i].end(); ++it) {
+                os << *it-seq.begin() << " ";
+                this_bin_fill++;
+            }
+            os << endl;
+            cum_fill += this_bin_fill;
+            cum_fill_sq += this_bin_fill*this_bin_fill;
+        }
+        if (this_bin_fill<MAXCOUNT) {
+            ++count[this_bin_fill];
+        } else {
+            ++count[MAXCOUNT];
+        }
+    }
+    unsigned fact=1;
+    double cum_prob=0;
+    double lambda=(1.0*seq.size())/hashtabsize;
+    for (int k=0; k<MAXCOUNT; ++k) {
+        if (k>0) fact *= k;
+        double prob=pow(lambda,k)/fact*exp(-lambda);
+        cum_prob += prob;
+        os << string(indent+delta_indent,' ') << "bins with " << k << " entries: "
+           << int(1000.0*count[k]/hashtabsize)/10.0 << "% (expected: "
+           << int(prob*1000)/10.0 << ")" << endl;
+    }
+    os << string(indent+delta_indent,' ') << "bins with more entries: "
+       << int(1000.0*count[MAXCOUNT]/hashtabsize)/10.0 << "% (expected: "
+       << int((1-cum_prob)*1000)/10.0 << ")" << endl;
+    
+    os << string(indent+delta_indent,' ') << "variance: "
+       << 1.0/hashtabsize*cum_fill_sq-(1.0/hashtabsize*cum_fill)*(1.0/hashtabsize*cum_fill)
+       << endl;
+    os << string(indent+delta_indent,' ') << "average fill: "
+       << (1.0*cum_fill)/hashtabsize
+       << " (should be equal to " << (1.0*seq.size())/hashtabsize << ")" << endl;
+#endif // def EXPAIRSEQ_USE_HASHTAB
+}
+
+bool expairseq::info(unsigned inf) const
+{
+    return inherited::info(inf);
+}
+
+unsigned expairseq::nops() const
 {
     if (overall_coeff.is_equal(default_overall_coeff())) {
         return seq.size();
@@ -175,7 +320,7 @@ int expairseq::nops() const
     return seq.size()+1;
 }
 
-ex expairseq::op(int const i) const
+ex expairseq::op(int i) const
 {
     if (unsigned(i)<seq.size()) {
         return recombine_pair_to_ex(seq[i]);
@@ -184,7 +329,7 @@ ex expairseq::op(int const i) const
     return overall_coeff;
 }
 
-ex & expairseq::let_op(int const i)
+ex & expairseq::let_op(int i)
 {
     throw(std::logic_error("let_op not defined for expairseq and derived classes (add,mul,...)"));
 }
@@ -216,7 +361,7 @@ ex expairseq::normal(lst &sym_lst, lst &repl_lst, int level) const
     return n.bp->basic::normal(sym_lst,repl_lst,level);
 }
 
-ex expairseq::subs(lst const & ls, lst const & lr) const
+ex expairseq::subs(const lst & ls, const lst & lr) const
 {
     epvector * vp=subschildren(ls,lr);
     if (vp==0) {
@@ -227,10 +372,18 @@ ex expairseq::subs(lst const & ls, lst const & lr) const
 
 // protected
 
-int expairseq::compare_same_type(basic const & other) const
+/** Implementation of ex::diff() for an expairseq. It differentiates all elements of the
+ *  sequence.
+ *  @see ex::diff */
+ex expairseq::derivative(const symbol & s) const
+{
+    return thisexpairseq(diffchildren(s),overall_coeff);
+}
+
+int expairseq::compare_same_type(const basic & other) const
 {
     GINAC_ASSERT(is_of_type(other, expairseq));
-    expairseq const & o=static_cast<expairseq const &>(const_cast<basic &>(other));
+    const expairseq & o=static_cast<const expairseq &>(const_cast<basic &>(other));
 
     int cmpval;
     
@@ -278,8 +431,8 @@ int expairseq::compare_same_type(basic const & other) const
     for (unsigned i=0; i<hashtabsize; ++i) {
         unsigned sz=hashtab[i].size();
         if (sz>0) {
-            epplist const & eppl1=hashtab[i];
-            epplist const & eppl2=o.hashtab[i];
+            const epplist & eppl1=hashtab[i];
+            const epplist & eppl2=o.hashtab[i];
             epplist::const_iterator it1=eppl1.begin();
             epplist::const_iterator it2=eppl2.begin();
             while (it1!=eppl1.end()) {
@@ -295,9 +448,9 @@ int expairseq::compare_same_type(basic const & other) const
 #endif // def EXPAIRSEQ_USE_HASHTAB
 }
 
-bool expairseq::is_equal_same_type(basic const & other) const
+bool expairseq::is_equal_same_type(const basic & other) const
 {
-    expairseq const & o=dynamic_cast<expairseq const &>(const_cast<basic &>(other));
+    const expairseq & o=dynamic_cast<const expairseq &>(const_cast<basic &>(other));
 
     // compare number of elements
     if (seq.size() != o.seq.size()) return false;
@@ -340,8 +493,8 @@ bool expairseq::is_equal_same_type(basic const & other) const
     for (unsigned i=0; i<hashtabsize; ++i) {
         unsigned sz=hashtab[i].size();
         if (sz>0) {
-            epplist const & eppl1=hashtab[i];
-            epplist const & eppl2=o.hashtab[i];
+            const epplist & eppl1=hashtab[i];
+            const epplist & eppl2=o.hashtab[i];
             epplist::const_iterator it1=eppl1.begin();
             epplist::const_iterator it2=eppl2.begin();
             while (it1!=eppl1.end()) {
@@ -399,31 +552,58 @@ ex expairseq::expand(unsigned options) const
 
 // protected
 
-ex expairseq::thisexpairseq(epvector const & v,ex const & oc) const
+ex expairseq::thisexpairseq(const epvector & v,const ex & oc) const
 {
     return expairseq(v,oc);
 }
 
-ex expairseq::thisexpairseq(epvector * vp, ex const & oc) const
+ex expairseq::thisexpairseq(epvector * vp, const ex & oc) const
 {
     return expairseq(vp,oc);
 }
 
-expair expairseq::split_ex_to_pair(ex const & e) const
+void expairseq::printpair(ostream & os, const expair & p, unsigned upper_precedence) const
 {
-    return expair(e,exONE());
+    os << "[[";
+    p.rest.bp->print(os,precedence);
+    os << ",";
+    p.coeff.bp->print(os,precedence);
+    os << "]]";
 }
 
-expair expairseq::combine_ex_with_coeff_to_pair(ex const & e,
-                                                ex const & c) const
+void expairseq::printseq(ostream & os, char delim, unsigned this_precedence,
+                         unsigned upper_precedence) const
+{
+    if (this_precedence<=upper_precedence) os << "(";
+    epvector::const_iterator it,it_last;
+    it_last=seq.end();
+    --it_last;
+    for (it=seq.begin(); it!=it_last; ++it) {
+        printpair(os,*it,this_precedence);
+        os << delim;
+    }
+    printpair(os,*it,this_precedence);
+    if (!overall_coeff.is_equal(default_overall_coeff())) {
+        os << delim << overall_coeff;
+    }
+    if (this_precedence<=upper_precedence) os << ")";
+}
+    
+expair expairseq::split_ex_to_pair(const ex & e) const
+{
+    return expair(e,_ex1());
+}
+
+expair expairseq::combine_ex_with_coeff_to_pair(const ex & e,
+                                                const ex & c) const
 {
     GINAC_ASSERT(is_ex_exactly_of_type(c,numeric));
 
     return expair(e,c);
 }
 
-expair expairseq::combine_pair_with_coeff_to_pair(expair const & p,
-                                                  ex const & c) const
+expair expairseq::combine_pair_with_coeff_to_pair(const expair & p,
+                                                  const ex & c) const
 {
     GINAC_ASSERT(is_ex_exactly_of_type(p.coeff,numeric));
     GINAC_ASSERT(is_ex_exactly_of_type(c,numeric));
@@ -431,7 +611,7 @@ expair expairseq::combine_pair_with_coeff_to_pair(expair const & p,
     return expair(p.rest,ex_to_numeric(p.coeff).mul_dyn(ex_to_numeric(c)));
 }
 
-ex expairseq::recombine_pair_to_ex(expair const & p) const
+ex expairseq::recombine_pair_to_ex(const expair & p) const
 {
     return lst(p.rest,p.coeff);
 }
@@ -443,17 +623,17 @@ bool expairseq::expair_needs_further_processing(epp it)
 
 ex expairseq::default_overall_coeff(void) const
 {
-    return exZERO();
+    return _ex0();
 }
 
-void expairseq::combine_overall_coeff(ex const & c)
+void expairseq::combine_overall_coeff(const ex & c)
 {
     GINAC_ASSERT(is_ex_exactly_of_type(overall_coeff,numeric));
     GINAC_ASSERT(is_ex_exactly_of_type(c,numeric));
     overall_coeff = ex_to_numeric(overall_coeff).add_dyn(ex_to_numeric(c));
 }
 
-void expairseq::combine_overall_coeff(ex const & c1, ex const & c2)
+void expairseq::combine_overall_coeff(const ex & c1, const ex & c2)
 {
     GINAC_ASSERT(is_ex_exactly_of_type(overall_coeff,numeric));
     GINAC_ASSERT(is_ex_exactly_of_type(c1,numeric));
@@ -462,7 +642,7 @@ void expairseq::combine_overall_coeff(ex const & c1, ex const & c2)
                         add_dyn(ex_to_numeric(c1).mul(ex_to_numeric(c2)));
 }
 
-bool expairseq::can_make_flat(expair const & p) const
+bool expairseq::can_make_flat(const expair & p) const
 {
     return true;
 }
@@ -472,7 +652,7 @@ bool expairseq::can_make_flat(expair const & p) const
 // non-virtual functions in this class
 //////////
 
-void expairseq::construct_from_2_ex_via_exvector(ex const & lh, ex const & rh)
+void expairseq::construct_from_2_ex_via_exvector(const ex & lh, const ex & rh)
 {
     exvector v;
     v.reserve(2);
@@ -485,7 +665,7 @@ void expairseq::construct_from_2_ex_via_exvector(ex const & lh, ex const & rh)
 #endif // def EXPAIRSEQ_USE_HASHTAB
 }
 
-void expairseq::construct_from_2_ex(ex const & lh, ex const & rh)
+void expairseq::construct_from_2_ex(const ex & lh, const ex & rh)
 {
     if (lh.bp->tinfo()==tinfo()) {
        if (rh.bp->tinfo()==tinfo()) {
@@ -575,8 +755,8 @@ void expairseq::construct_from_2_ex(ex const & lh, ex const & rh)
     }
 }
 
-void expairseq::construct_from_2_expairseq(expairseq const & s1,
-                                           expairseq const & s2)
+void expairseq::construct_from_2_expairseq(const expairseq & s1,
+                                           const expairseq & s2)
 {
     combine_overall_coeff(s1.overall_coeff);
     combine_overall_coeff(s2.overall_coeff);
@@ -594,7 +774,7 @@ void expairseq::construct_from_2_expairseq(expairseq const & s1,
         int cmpval=(*first1).rest.compare((*first2).rest);
         if (cmpval==0) {
             // combine terms
-            numeric const & newcoeff=ex_to_numeric((*first1).coeff).
+            const numeric & newcoeff=ex_to_numeric((*first1).coeff).
                                      add(ex_to_numeric((*first2).coeff));
             if (!newcoeff.is_zero()) {
                 seq.push_back(expair((*first1).rest,newcoeff));
@@ -629,8 +809,8 @@ void expairseq::construct_from_2_expairseq(expairseq const & s1,
     }
 }
 
-void expairseq::construct_from_expairseq_ex(expairseq const & s,
-                                            ex const & e)
+void expairseq::construct_from_expairseq_ex(const expairseq & s,
+                                            const ex & e)
 {
     combine_overall_coeff(s.overall_coeff);
     if (is_ex_exactly_of_type(e,numeric)) {
@@ -653,7 +833,7 @@ void expairseq::construct_from_expairseq_ex(expairseq const & s,
         int cmpval=(*first).rest.compare(p.rest);
         if (cmpval==0) {
             // combine terms
-            numeric const & newcoeff=ex_to_numeric((*first).coeff).
+            const numeric & newcoeff=ex_to_numeric((*first).coeff).
                                      add(ex_to_numeric(p.coeff));
             if (!newcoeff.is_zero()) {
                 seq.push_back(expair((*first).rest,newcoeff));
@@ -692,7 +872,7 @@ void expairseq::construct_from_expairseq_ex(expairseq const & s,
     }
 }
 
-void expairseq::construct_from_exvector(exvector const & v)
+void expairseq::construct_from_exvector(const exvector & v)
 {
     // simplifications: +(a,+(b,c),d) -> +(a,b,c,d) (associativity)
     //                  +(d,b,c,a) -> +(a,b,c,d) (canonicalization)
@@ -708,7 +888,7 @@ void expairseq::construct_from_exvector(exvector const & v)
 #endif // def EXPAIRSEQ_USE_HASHTAB
 }
 
-void expairseq::construct_from_epvector(epvector const & v)
+void expairseq::construct_from_epvector(const epvector & v)
 {
     // simplifications: +(a,+(b,c),d) -> +(a,b,c,d) (associativity)
     //                  +(d,b,c,a) -> +(a,b,c,d) (canonicalization)
@@ -726,7 +906,7 @@ void expairseq::construct_from_epvector(epvector const & v)
 
 #include <iostream>
 
-void expairseq::make_flat(exvector const & v)
+void expairseq::make_flat(const exvector & v)
 {
     exvector::const_iterator cit, citend = v.end();
 
@@ -750,7 +930,7 @@ void expairseq::make_flat(exvector const & v)
     cit=v.begin();
     while (cit!=citend) {
         if (cit->bp->tinfo()==tinfo()) {
-            expairseq const & subseqref=ex_to_expairseq(*cit);
+            const expairseq & subseqref=ex_to_expairseq(*cit);
             combine_overall_coeff(subseqref.overall_coeff);
             epvector::const_iterator cit_s=subseqref.seq.begin();
             while (cit_s!=subseqref.seq.end()) {
@@ -777,7 +957,7 @@ void expairseq::make_flat(exvector const & v)
     */
 }
 
-void expairseq::make_flat(epvector const & v)
+void expairseq::make_flat(const epvector & v)
 {
     epvector::const_iterator cit, citend = v.end();
 
@@ -801,7 +981,7 @@ void expairseq::make_flat(epvector const & v)
     cit=v.begin();
     while (cit!=citend) {
         if ((cit->rest.bp->tinfo()==tinfo())&&can_make_flat(*cit)) {
-            expairseq const & subseqref=ex_to_expairseq((*cit).rest);
+            const expairseq & subseqref=ex_to_expairseq((*cit).rest);
             combine_overall_coeff(ex_to_numeric(subseqref.overall_coeff),
                                   ex_to_numeric((*cit).coeff));
             epvector::const_iterator cit_s=subseqref.seq.begin();
@@ -1014,7 +1194,7 @@ unsigned expairseq::calc_hashtabsize(unsigned sz) const
     return size;
 }
 
-unsigned expairseq::calc_hashindex(ex const & e) const
+unsigned expairseq::calc_hashindex(const ex & e) const
 {
     // calculate hashindex
     unsigned hash=e.gethash();
@@ -1186,7 +1366,7 @@ void expairseq::drop_coeff_0_terms(epvector::iterator & first_numeric,
         if (!touched[i]) {
             ++current;
             ++i;
-        } else if (!ex_to_numeric((*current).coeff).is_equal(numZERO())) {
+        } else if (!ex_to_numeric((*current).coeff).is_equal(_num0())) {
             ++current;
             ++i;
         } else {
@@ -1228,7 +1408,7 @@ void expairseq::drop_coeff_0_terms(epvector::iterator & first_numeric,
 bool expairseq::has_coeff_0(void) const
 {
     for (epvector::const_iterator cit=seq.begin(); cit!=seq.end(); ++cit) {
-        if ((*cit).coeff.is_equal(exZERO())) {
+        if ((*cit).coeff.is_equal(_ex0())) {
             return true;
         }
     }
@@ -1368,7 +1548,7 @@ epvector * expairseq::expandchildren(unsigned options) const
     epvector::const_iterator last=seq.end();
     epvector::const_iterator cit=seq.begin();
     while (cit!=last) {
-        ex const & expanded_ex=(*cit).rest.expand(options);
+        const ex & expanded_ex=(*cit).rest.expand(options);
         if (!are_ex_trivially_equal((*cit).rest,expanded_ex)) {
 
             // something changed, copy seq, eval and return it
@@ -1416,10 +1596,10 @@ epvector * expairseq::evalchildren(int level) const
     epvector::const_iterator last=seq.end();
     epvector::const_iterator cit=seq.begin();
     while (cit!=last) {
-        ex const & evaled_ex=(*cit).rest.eval(level);
+        const ex & evaled_ex=(*cit).rest.eval(level);
         if (!are_ex_trivially_equal((*cit).rest,evaled_ex)) {
 
-	    // something changed, copy seq, eval and return it
+            // something changed, copy seq, eval and return it
             epvector *s=new epvector;
             s->reserve(seq.size());
 
@@ -1485,7 +1665,7 @@ epvector expairseq::normalchildren(int level) const
     return s;
 }
 
-epvector expairseq::diffchildren(symbol const & y) const
+epvector expairseq::diffchildren(const symbol & y) const
 {
     epvector s;
     s.reserve(seq.size());
@@ -1497,22 +1677,23 @@ epvector expairseq::diffchildren(symbol const & y) const
     return s;
 }
 
-epvector * expairseq::subschildren(lst const & ls, lst const & lr) const
+epvector * expairseq::subschildren(const lst & ls, const lst & lr) const
 {
     // returns a NULL pointer if nothing had to be substituted
     // returns a pointer to a newly created epvector otherwise
     // (which has to be deleted somewhere else)
-
+    GINAC_ASSERT(ls.nops()==lr.nops());
+    
     epvector::const_iterator last=seq.end();
     epvector::const_iterator cit=seq.begin();
     while (cit!=last) {
-        ex const & subsed_ex=(*cit).rest.subs(ls,lr);
+        const ex & subsed_ex=(*cit).rest.subs(ls,lr);
         if (!are_ex_trivially_equal((*cit).rest,subsed_ex)) {
-
+            
             // something changed, copy seq, subs and return it
             epvector *s=new epvector;
             s->reserve(seq.size());
-
+            
             // copy parts of seq which are known not to have changed
             epvector::const_iterator cit2=seq.begin();
             while (cit2!=cit) {
@@ -1528,7 +1709,7 @@ epvector * expairseq::subschildren(lst const & ls, lst const & lr) const
                 s->push_back(combine_ex_with_coeff_to_pair((*cit2).rest.subs(ls,lr),
                                                            (*cit2).coeff));
                 ++cit2;
-	    }
+            }
             return s;
         }
         ++cit;
@@ -1537,75 +1718,6 @@ epvector * expairseq::subschildren(lst const & ls, lst const & lr) const
     return 0; // nothing has changed
 }
 
-/*
-epvector expairseq::subschildren(lst const & ls, lst const & lr) const
-{
-    epvector s;
-    s.reserve(seq.size());
-
-    for (epvector::const_iterator it=seq.begin(); it!=seq.end(); ++it) {
-        s.push_back(split_ex_to_pair((*it).rest.subs(ls,lr),(*it).coeff));
-    }
-    return s;
-}
-*/
-
-/*
-void expairseq::sort(epviter first, epviter last, expair_is_less comp)
-{
-    if (first != last) {
-        introsort_loop(first, last, lg(last - first) * 2, comp);
-        __final_insertion_sort(first, last, comp);
-    }
-}
-
-ptrdiff_t expairseq::lg(ptrdiff_t n)
-{
-    ptrdiff_t k;
-    for (k = 0; n > 1; n >>= 1) ++k;
-    return k;
-}
-
-void expairseq::introsort_loop(epviter first, epviter last,
-                               ptrdiff_t depth_limit, expair_is_less comp)
-{
-    while (last - first > stl_threshold) {
-        if (depth_limit == 0) {
-            partial_sort(first, last, last, comp);
-            return;
-        }
-        --depth_limit;
-        epviter cut = unguarded_partition(first, last,
-                      expair(__median(*first, *(first + (last - first)/2),
-                      *(last - 1), comp)), comp);
-        introsort_loop(cut, last, depth_limit, comp);
-        last = cut;
-    }
-}
-
-epviter expairseq::unguarded_partition(epviter first, epviter last, 
-                                       expair pivot, expair_is_less comp)
-{
-    while (1) {
-        while (comp(*first, pivot)) ++first;
-        --last;
-        while (comp(pivot, *last)) --last;
-        if (!(first < last)) return first;
-        iter_swap(first, last);
-        ++first;
-    }
-}
-
-void expairseq::partial_sort(epviter first, epviter middle, epviter last,
-                             expair_is_less comp) {
-  make_heap(first, middle, comp);
-  for (RandomAccessIterator i = middle; i < last; ++i)
-    if (comp(*i, *first))
-      __pop_heap(first, middle, i, T(*i), comp, distance_type(first));
-  sort_heap(first, middle, comp);
-}
-*/
-
 //////////
 // static member variables
 //////////
@@ -1625,6 +1737,8 @@ unsigned expairseq::hashtabfactor=1;
 //////////
 
 const expairseq some_expairseq;
-type_info const & typeid_expairseq=typeid(some_expairseq);
+const type_info & typeid_expairseq=typeid(some_expairseq);
 
+#ifndef NO_NAMESPACE_GINAC
 } // namespace GiNaC
+#endif // ndef NO_NAMESPACE_GINAC
