@@ -3,7 +3,7 @@
  *  Implementation of GiNaC's symbolic exponentiation (basis^exponent). */
 
 /*
- *  GiNaC Copyright (C) 1999 Johannes Gutenberg University Mainz, Germany
+ *  GiNaC Copyright (C) 1999-2000 Johannes Gutenberg University Mainz, Germany
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -29,11 +29,18 @@
 #include "add.h"
 #include "mul.h"
 #include "numeric.h"
+#include "inifcns.h"
 #include "relational.h"
 #include "symbol.h"
+#include "archive.h"
 #include "debugmsg.h"
+#include "utils.h"
 
+#ifndef NO_NAMESPACE_GINAC
 namespace GiNaC {
+#endif // ndef NO_NAMESPACE_GINAC
+
+GINAC_IMPLEMENT_REGISTERED_CLASS(power, basic)
 
 typedef vector<int> intvector;
 
@@ -54,13 +61,13 @@ power::~power()
     destroy(0);
 }
 
-power::power(power const & other)
+power::power(const power & other)
 {
     debugmsg("power copy constructor",LOGLEVEL_CONSTRUCT);
     copy(other);
 }
 
-power const & power::operator=(power const & other)
+const power & power::operator=(const power & other)
 {
     debugmsg("power operator=",LOGLEVEL_ASSIGNMENT);
     if (this != &other) {
@@ -72,16 +79,16 @@ power const & power::operator=(power const & other)
 
 // protected
 
-void power::copy(power const & other)
+void power::copy(const power & other)
 {
-    basic::copy(other);
+    inherited::copy(other);
     basis=other.basis;
     exponent=other.exponent;
 }
 
 void power::destroy(bool call_parent)
 {
-    if (call_parent) basic::destroy(call_parent);
+    if (call_parent) inherited::destroy(call_parent);
 }
 
 //////////
@@ -90,18 +97,44 @@ void power::destroy(bool call_parent)
 
 // public
 
-power::power(ex const & lh, ex const & rh) : basic(TINFO_power), basis(lh), exponent(rh)
+power::power(const ex & lh, const ex & rh) : basic(TINFO_power), basis(lh), exponent(rh)
 {
     debugmsg("power constructor from ex,ex",LOGLEVEL_CONSTRUCT);
     GINAC_ASSERT(basis.return_type()==return_types::commutative);
 }
 
-power::power(ex const & lh, numeric const & rh) : basic(TINFO_power), basis(lh), exponent(rh)
+power::power(const ex & lh, const numeric & rh) : basic(TINFO_power), basis(lh), exponent(rh)
 {
     debugmsg("power constructor from ex,numeric",LOGLEVEL_CONSTRUCT);
     GINAC_ASSERT(basis.return_type()==return_types::commutative);
 }
 
+//////////
+// archiving
+//////////
+
+/** Construct object from archive_node. */
+power::power(const archive_node &n, const lst &sym_lst) : inherited(n, sym_lst)
+{
+    debugmsg("power constructor from archive_node", LOGLEVEL_CONSTRUCT);
+    n.find_ex("basis", basis, sym_lst);
+    n.find_ex("exponent", exponent, sym_lst);
+}
+
+/** Unarchive the object. */
+ex power::unarchive(const archive_node &n, const lst &sym_lst)
+{
+    return (new power(n, sym_lst))->setflag(status_flags::dynallocated);
+}
+
+/** Archive the object. */
+void power::archive(archive_node &n) const
+{
+    inherited::archive(n);
+    n.add_ex("basis", basis);
+    n.add_ex("exponent", exponent);
+}
+
 //////////
 // functions overriding virtual functions from bases classes
 //////////
@@ -114,23 +147,128 @@ basic * power::duplicate() const
     return new power(*this);
 }
 
+void power::print(ostream & os, unsigned upper_precedence) const
+{
+    debugmsg("power print",LOGLEVEL_PRINT);
+    if (exponent.is_equal(_ex1_2())) {
+        os << "sqrt(" << basis << ")";
+    } else {
+        if (precedence<=upper_precedence) os << "(";
+        basis.print(os,precedence);
+        os << "^";
+        exponent.print(os,precedence);
+        if (precedence<=upper_precedence) os << ")";
+    }
+}
+
+void power::printraw(ostream & os) const
+{
+    debugmsg("power printraw",LOGLEVEL_PRINT);
+
+    os << "power(";
+    basis.printraw(os);
+    os << ",";
+    exponent.printraw(os);
+    os << ",hash=" << hashvalue << ",flags=" << flags << ")";
+}
+
+void power::printtree(ostream & os, unsigned indent) const
+{
+    debugmsg("power printtree",LOGLEVEL_PRINT);
+
+    os << string(indent,' ') << "power: "
+       << "hash=" << hashvalue << " (0x" << hex << hashvalue << dec << ")"
+       << ", flags=" << flags << endl;
+    basis.printtree(os,indent+delta_indent);
+    exponent.printtree(os,indent+delta_indent);
+}
+
+static void print_sym_pow(ostream & os, unsigned type, const symbol &x, int exp)
+{
+    // Optimal output of integer powers of symbols to aid compiler CSE
+    if (exp == 1) {
+        x.printcsrc(os, type, 0);
+    } else if (exp == 2) {
+        x.printcsrc(os, type, 0);
+        os << "*";
+        x.printcsrc(os, type, 0);
+    } else if (exp & 1) {
+        x.printcsrc(os, 0);
+        os << "*";
+        print_sym_pow(os, type, x, exp-1);
+    } else {
+        os << "(";
+        print_sym_pow(os, type, x, exp >> 1);
+        os << ")*(";
+        print_sym_pow(os, type, x, exp >> 1);
+        os << ")";
+    }
+}
+
+void power::printcsrc(ostream & os, unsigned type, unsigned upper_precedence) const
+{
+    debugmsg("power print csrc", LOGLEVEL_PRINT);
+    
+    // Integer powers of symbols are printed in a special, optimized way
+    if (exponent.info(info_flags::integer) &&
+        (is_ex_exactly_of_type(basis, symbol) ||
+         is_ex_exactly_of_type(basis, constant))) {
+        int exp = ex_to_numeric(exponent).to_int();
+        if (exp > 0)
+            os << "(";
+        else {
+            exp = -exp;
+            if (type == csrc_types::ctype_cl_N)
+                os << "recip(";
+            else
+                os << "1.0/(";
+        }
+        print_sym_pow(os, type, static_cast<const symbol &>(*basis.bp), exp);
+        os << ")";
+
+    // <expr>^-1 is printed as "1.0/<expr>" or with the recip() function of CLN
+    } else if (exponent.compare(_num_1()) == 0) {
+        if (type == csrc_types::ctype_cl_N)
+            os << "recip(";
+        else
+            os << "1.0/(";
+        basis.bp->printcsrc(os, type, 0);
+        os << ")";
+
+    // Otherwise, use the pow() or expt() (CLN) functions
+    } else {
+        if (type == csrc_types::ctype_cl_N)
+            os << "expt(";
+        else
+            os << "pow(";
+        basis.bp->printcsrc(os, type, 0);
+        os << ",";
+        exponent.bp->printcsrc(os, type, 0);
+        os << ")";
+    }
+}
+
 bool power::info(unsigned inf) const
 {
-    if (inf==info_flags::polynomial || inf==info_flags::integer_polynomial || inf==info_flags::rational_polynomial) {
+    if (inf==info_flags::polynomial ||
+        inf==info_flags::integer_polynomial ||
+        inf==info_flags::cinteger_polynomial ||
+        inf==info_flags::rational_polynomial ||
+        inf==info_flags::crational_polynomial) {
         return exponent.info(info_flags::nonnegint);
     } else if (inf==info_flags::rational_function) {
         return exponent.info(info_flags::integer);
     } else {
-        return basic::info(inf);
+        return inherited::info(inf);
     }
 }
 
-int power::nops() const
+unsigned power::nops() const
 {
     return 2;
 }
 
-ex & power::let_op(int const i)
+ex & power::let_op(int i)
 {
     GINAC_ASSERT(i>=0);
     GINAC_ASSERT(i<2);
@@ -138,10 +276,10 @@ ex & power::let_op(int const i)
     return i==0 ? basis : exponent;
 }
 
-int power::degree(symbol const & s) const
+int power::degree(const symbol & s) const
 {
     if (is_exactly_of_type(*exponent.bp,numeric)) {
-	if ((*basis.bp).compare(s)==0)
+        if ((*basis.bp).compare(s)==0)
             return ex_to_numeric(exponent).to_int();
         else
             return basis.degree(s) * ex_to_numeric(exponent).to_int();
@@ -149,10 +287,10 @@ int power::degree(symbol const & s) const
     return 0;
 }
 
-int power::ldegree(symbol const & s) const 
+int power::ldegree(const symbol & s) const 
 {
     if (is_exactly_of_type(*exponent.bp,numeric)) {
-	if ((*basis.bp).compare(s)==0)
+        if ((*basis.bp).compare(s)==0)
             return ex_to_numeric(exponent).to_int();
         else
             return basis.ldegree(s) * ex_to_numeric(exponent).to_int();
@@ -160,21 +298,21 @@ int power::ldegree(symbol const & s) const
     return 0;
 }
 
-ex power::coeff(symbol const & s, int const n) const
+ex power::coeff(const symbol & s, int n) const
 {
     if ((*basis.bp).compare(s)!=0) {
         // basis not equal to s
         if (n==0) {
             return *this;
         } else {
-            return exZERO();
+            return _ex0();
         }
     } else if (is_exactly_of_type(*exponent.bp,numeric)&&
-               (static_cast<numeric const &>(*exponent.bp).compare(numeric(n))==0)) {
-        return exONE();
+               (static_cast<const numeric &>(*exponent.bp).compare(numeric(n))==0)) {
+        return _ex1();
     }
 
-    return exZERO();
+    return _ex0();
 }
 
 ex power::eval(int level) const
@@ -197,8 +335,8 @@ ex power::eval(int level) const
         throw(std::runtime_error("max recursion level reached"));
     }
     
-    ex const & ebasis    = level==1 ? basis    : basis.eval(level-1);
-    ex const & eexponent = level==1 ? exponent : exponent.eval(level-1);
+    const ex & ebasis    = level==1 ? basis    : basis.eval(level-1);
+    const ex & eexponent = level==1 ? exponent : exponent.eval(level-1);
 
     bool basis_is_numerical=0;
     bool exponent_is_numerical=0;
@@ -216,10 +354,13 @@ ex power::eval(int level) const
 
     // ^(x,0) -> 1 (0^0 also handled here)
     if (eexponent.is_zero())
-        return exONE();
+        if (ebasis.is_zero())
+            throw (std::domain_error("power::eval(): pow(0,0) is undefined"));
+        else
+            return _ex1();
 
     // ^(x,1) -> x
-    if (eexponent.is_equal(exONE()))
+    if (eexponent.is_equal(_ex1()))
         return ebasis;
 
     // ^(0,x) -> 0 (except if x is real and negative)
@@ -227,27 +368,27 @@ ex power::eval(int level) const
         if (exponent_is_numerical && num_exponent->is_negative()) {
             throw(std::overflow_error("power::eval(): division by zero"));
         } else
-            return exZERO();
+            return _ex0();
     }
 
     // ^(1,x) -> 1
-    if (ebasis.is_equal(exONE()))
-        return exONE();
+    if (ebasis.is_equal(_ex1()))
+        return _ex1();
 
     if (basis_is_numerical && exponent_is_numerical) {
         // ^(c1,c2) -> c1^c2 (c1, c2 numeric(),
         // except if c1,c2 are rational, but c1^c2 is not)
-        bool basis_is_rational = num_basis->is_rational();
-        bool exponent_is_rational = num_exponent->is_rational();
+        bool basis_is_crational = num_basis->is_crational();
+        bool exponent_is_crational = num_exponent->is_crational();
         numeric res = (*num_basis).power(*num_exponent);
         
-        if ((!basis_is_rational || !exponent_is_rational)
-            || res.is_rational()) {
+        if ((!basis_is_crational || !exponent_is_crational)
+            || res.is_crational()) {
             return res;
         }
         GINAC_ASSERT(!num_exponent->is_integer());  // has been handled by now
         // ^(c1,n/m) -> *(c1^q,c1^(n/m-q)), 0<(n/m-h)<1, q integer
-        if (basis_is_rational && exponent_is_rational
+        if (basis_is_crational && exponent_is_crational
             && num_exponent->is_real()
             && !num_exponent->is_integer()) {
             numeric r, q, n, m;
@@ -256,14 +397,14 @@ ex power::eval(int level) const
             q = iquo(n, m, r);
             if (r.is_negative()) {
                 r = r.add(m);
-                q = q.sub(numONE());
+                q = q.sub(_num1());
             }
             if (q.is_zero())  // the exponent was in the allowed range 0<(n/m)<1
                 return this->hold();
             else {
                 epvector res(2);
                 res.push_back(expair(ebasis,r.div(m)));
-                res.push_back(expair(ex(num_basis->power(q)),exONE()));
+                res.push_back(expair(ex(num_basis->power(q)),_ex1()));
                 return (new mul(res))->setflag(status_flags::dynallocated | status_flags::evaluated);
                 /*return mul(num_basis->power(q),
                            power(ex(*num_basis),ex(r.div(m)))).hold();
@@ -279,11 +420,11 @@ ex power::eval(int level) const
     // (c1, c2 numeric(), c2 integer or -1 < c1 <= 1,
     // case c1=1 should not happen, see below!)
     if (exponent_is_numerical && is_ex_exactly_of_type(ebasis,power)) {
-        power const & sub_power=ex_to_power(ebasis);
-        ex const & sub_basis=sub_power.basis;
-        ex const & sub_exponent=sub_power.exponent;
+        const power & sub_power=ex_to_power(ebasis);
+        const ex & sub_basis=sub_power.basis;
+        const ex & sub_exponent=sub_power.exponent;
         if (is_ex_exactly_of_type(sub_exponent,numeric)) {
-            numeric const & num_sub_exponent=ex_to_numeric(sub_exponent);
+            const numeric & num_sub_exponent=ex_to_numeric(sub_exponent);
             GINAC_ASSERT(num_sub_exponent!=numeric(1));
             if (num_exponent->is_integer() || abs(num_sub_exponent)<1) {
                 return power(sub_basis,num_sub_exponent.mul(*num_exponent));
@@ -301,23 +442,23 @@ ex power::eval(int level) const
     // ^(*(...,x,c1),c2) -> ^(*(...,x;-1),c2)*(-c1)^c2 (c1, c2 numeric(), c1<0)
     if (exponent_is_numerical && is_ex_exactly_of_type(ebasis,mul)) {
         GINAC_ASSERT(!num_exponent->is_integer()); // should have been handled above
-        mul const & mulref=ex_to_mul(ebasis);
-        if (!mulref.overall_coeff.is_equal(exONE())) {
-            numeric const & num_coeff=ex_to_numeric(mulref.overall_coeff);
+        const mul & mulref=ex_to_mul(ebasis);
+        if (!mulref.overall_coeff.is_equal(_ex1())) {
+            const numeric & num_coeff=ex_to_numeric(mulref.overall_coeff);
             if (num_coeff.is_real()) {
                 if (num_coeff.is_positive()>0) {
                     mul * mulp=new mul(mulref);
-                    mulp->overall_coeff=exONE();
+                    mulp->overall_coeff=_ex1();
                     mulp->clearflag(status_flags::evaluated);
                     mulp->clearflag(status_flags::hash_calculated);
                     return (new mul(power(*mulp,exponent),
                                     power(num_coeff,*num_exponent)))->
                         setflag(status_flags::dynallocated);
                 } else {
-                    GINAC_ASSERT(num_coeff.compare(numZERO())<0);
-                    if (num_coeff.compare(numMINUSONE())!=0) {
+                    GINAC_ASSERT(num_coeff.compare(_num0())<0);
+                    if (num_coeff.compare(_num_1())!=0) {
                         mul * mulp=new mul(mulref);
-                        mulp->overall_coeff=exMINUSONE();
+                        mulp->overall_coeff=_ex_1();
                         mulp->clearflag(status_flags::evaluated);
                         mulp->clearflag(status_flags::hash_calculated);
                         return (new mul(power(*mulp,exponent),
@@ -357,10 +498,10 @@ ex power::evalf(int level) const
     return power(ebasis,eexponent);
 }
 
-ex power::subs(lst const & ls, lst const & lr) const
+ex power::subs(const lst & ls, const lst & lr) const
 {
-    ex const & subsed_basis=basis.subs(ls,lr);
-    ex const & subsed_exponent=exponent.subs(ls,lr);
+    const ex & subsed_basis=basis.subs(ls,lr);
+    const ex & subsed_exponent=exponent.subs(ls,lr);
 
     if (are_ex_trivially_equal(basis,subsed_basis)&&
         are_ex_trivially_equal(exponent,subsed_exponent)) {
@@ -370,17 +511,32 @@ ex power::subs(lst const & ls, lst const & lr) const
     return power(subsed_basis, subsed_exponent);
 }
 
-ex power::simplify_ncmul(exvector const & v) const
+ex power::simplify_ncmul(const exvector & v) const
 {
-    return basic::simplify_ncmul(v);
+    return inherited::simplify_ncmul(v);
 }
 
 // protected
 
-int power::compare_same_type(basic const & other) const
+/** Implementation of ex::diff() for a power.
+ *  @see ex::diff */
+ex power::derivative(const symbol & s) const
+{
+    if (exponent.info(info_flags::real)) {
+        // D(b^r) = r * b^(r-1) * D(b) (faster than the formula below)
+        return mul(mul(exponent, power(basis, exponent - _ex1())), basis.diff(s));
+    } else {
+        // D(b^e) = b^e * (D(e)*ln(b) + e*D(b)/b)
+        return mul(power(basis, exponent),
+                   add(mul(exponent.diff(s), log(basis)),
+                       mul(mul(exponent, basis.diff(s)), power(basis, -1))));
+    }
+}
+
+int power::compare_same_type(const basic & other) const
 {
     GINAC_ASSERT(is_exactly_of_type(other, power));
-    power const & o=static_cast<power const &>(const_cast<basic &>(other));
+    const power & o=static_cast<const power &>(const_cast<basic &>(other));
 
     int cmpval;
     cmpval=basis.compare(o.basis);
@@ -415,7 +571,7 @@ ex power::expand(unsigned options) const
     }
 
     // integer numeric exponent
-    numeric const & num_exponent=ex_to_numeric(exponent);
+    const numeric & num_exponent=ex_to_numeric(exponent);
     int int_exponent = num_exponent.to_int();
 
     if (int_exponent > 0 && is_ex_exactly_of_type(expanded_basis,add)) {
@@ -445,7 +601,7 @@ ex power::expand(unsigned options) const
 // non-virtual functions in this class
 //////////
 
-ex power::expand_add(add const & a, int const n) const
+ex power::expand_add(const add & a, int n) const
 {
     // expand a^n where a is an add and n is an integer
 
@@ -471,7 +627,7 @@ ex power::expand_add(add const & a, int const n) const
         exvector term;
         term.reserve(m+1);
         for (l=0; l<m-1; l++) {
-            ex const & b=a.op(l);
+            const ex & b=a.op(l);
             GINAC_ASSERT(!is_ex_exactly_of_type(b,add));
             GINAC_ASSERT(!is_ex_exactly_of_type(b,power)||
                    !is_ex_exactly_of_type(ex_to_power(b).exponent,numeric)||
@@ -483,7 +639,7 @@ ex power::expand_add(add const & a, int const n) const
             }
         }
 
-        ex const & b=a.op(l);
+        const ex & b=a.op(l);
         GINAC_ASSERT(!is_ex_exactly_of_type(b,add));
         GINAC_ASSERT(!is_ex_exactly_of_type(b,power)||
                !is_ex_exactly_of_type(ex_to_power(b).exponent,numeric)||
@@ -541,42 +697,7 @@ ex power::expand_add(add const & a, int const n) const
     return (new add(sum))->setflag(status_flags::dynallocated);
 }
 
-/*
-ex power::expand_add_2(add const & a) const
-{
-    // special case: expand a^2 where a is an add
-
-    epvector sum;
-    sum.reserve((a.seq.size()*(a.seq.size()+1))/2);
-    epvector::const_iterator last=a.seq.end();
-
-    for (epvector::const_iterator cit0=a.seq.begin(); cit0!=last; ++cit0) {
-        ex const & b=a.recombine_pair_to_ex(*cit0);
-        GINAC_ASSERT(!is_ex_exactly_of_type(b,add));
-        GINAC_ASSERT(!is_ex_exactly_of_type(b,power)||
-               !is_ex_exactly_of_type(ex_to_power(b).exponent,numeric)||
-               !ex_to_numeric(ex_to_power(b).exponent).is_pos_integer());
-        if (is_ex_exactly_of_type(b,mul)) {
-            sum.push_back(a.split_ex_to_pair(expand_mul(ex_to_mul(b),numTWO())));
-        } else {
-            sum.push_back(a.split_ex_to_pair((new power(b,exTWO()))->
-                                              setflag(status_flags::dynallocated)));
-        }
-        for (epvector::const_iterator cit1=cit0+1; cit1!=last; ++cit1) {
-            sum.push_back(a.split_ex_to_pair((new mul(a.recombine_pair_to_ex(*cit0),
-                                                      a.recombine_pair_to_ex(*cit1)))->
-                                              setflag(status_flags::dynallocated),
-                                             exTWO()));
-        }
-    }
-
-    GINAC_ASSERT(sum.size()==(a.seq.size()*(a.seq.size()+1))/2);
-
-    return (new add(sum))->setflag(status_flags::dynallocated);
-}
-*/
-
-ex power::expand_add_2(add const & a) const
+ex power::expand_add_2(const add & a) const
 {
     // special case: expand a^2 where a is an add
 
@@ -588,8 +709,8 @@ ex power::expand_add_2(add const & a) const
     // power(+(x,...,z;c),2)=power(+(x,...,z;0),2)+2*c*+(x,...,z;0)+c*c
     // first part: ignore overall_coeff and expand other terms
     for (epvector::const_iterator cit0=a.seq.begin(); cit0!=last; ++cit0) {
-        ex const & r=(*cit0).rest;
-        ex const & c=(*cit0).coeff;
+        const ex & r=(*cit0).rest;
+        const ex & c=(*cit0).coeff;
         
         GINAC_ASSERT(!is_ex_exactly_of_type(r,add));
         GINAC_ASSERT(!is_ex_exactly_of_type(r,power)||
@@ -599,39 +720,39 @@ ex power::expand_add_2(add const & a) const
                !is_ex_exactly_of_type(ex_to_power(r).basis,mul)||
                !is_ex_exactly_of_type(ex_to_power(r).basis,power));
 
-        if (are_ex_trivially_equal(c,exONE())) {
+        if (are_ex_trivially_equal(c,_ex1())) {
             if (is_ex_exactly_of_type(r,mul)) {
-                sum.push_back(expair(expand_mul(ex_to_mul(r),numTWO()),exONE()));
+                sum.push_back(expair(expand_mul(ex_to_mul(r),_num2()),_ex1()));
             } else {
-                sum.push_back(expair((new power(r,exTWO()))->setflag(status_flags::dynallocated),
-                                     exONE()));
+                sum.push_back(expair((new power(r,_ex2()))->setflag(status_flags::dynallocated),
+                                     _ex1()));
             }
         } else {
             if (is_ex_exactly_of_type(r,mul)) {
-                sum.push_back(expair(expand_mul(ex_to_mul(r),numTWO()),
-                                     ex_to_numeric(c).power_dyn(numTWO())));
+                sum.push_back(expair(expand_mul(ex_to_mul(r),_num2()),
+                                     ex_to_numeric(c).power_dyn(_num2())));
             } else {
-                sum.push_back(expair((new power(r,exTWO()))->setflag(status_flags::dynallocated),
-                                     ex_to_numeric(c).power_dyn(numTWO())));
+                sum.push_back(expair((new power(r,_ex2()))->setflag(status_flags::dynallocated),
+                                     ex_to_numeric(c).power_dyn(_num2())));
             }
         }
             
         for (epvector::const_iterator cit1=cit0+1; cit1!=last; ++cit1) {
-            ex const & r1=(*cit1).rest;
-            ex const & c1=(*cit1).coeff;
+            const ex & r1=(*cit1).rest;
+            const ex & c1=(*cit1).coeff;
             sum.push_back(a.combine_ex_with_coeff_to_pair((new mul(r,r1))->setflag(status_flags::dynallocated),
-                                                          numTWO().mul(ex_to_numeric(c)).mul_dyn(ex_to_numeric(c1))));
+                                                          _num2().mul(ex_to_numeric(c)).mul_dyn(ex_to_numeric(c1))));
         }
     }
 
     GINAC_ASSERT(sum.size()==(a.seq.size()*(a.seq.size()+1))/2);
 
     // second part: add terms coming from overall_factor (if != 0)
-    if (!a.overall_coeff.is_equal(exZERO())) {
+    if (!a.overall_coeff.is_equal(_ex0())) {
         for (epvector::const_iterator cit=a.seq.begin(); cit!=a.seq.end(); ++cit) {
-            sum.push_back(a.combine_pair_with_coeff_to_pair(*cit,ex_to_numeric(a.overall_coeff).mul_dyn(numTWO())));
+            sum.push_back(a.combine_pair_with_coeff_to_pair(*cit,ex_to_numeric(a.overall_coeff).mul_dyn(_num2())));
         }
-        sum.push_back(expair(ex_to_numeric(a.overall_coeff).power_dyn(numTWO()),exONE()));
+        sum.push_back(expair(ex_to_numeric(a.overall_coeff).power_dyn(_num2()),_ex1()));
     }
         
     GINAC_ASSERT(sum.size()==(a_nops*(a_nops+1))/2);
@@ -639,12 +760,12 @@ ex power::expand_add_2(add const & a) const
     return (new add(sum))->setflag(status_flags::dynallocated);
 }
 
-ex power::expand_mul(mul const & m, numeric const & n) const
+ex power::expand_mul(const mul & m, const numeric & n) const
 {
     // expand m^n where m is a mul and n is and integer
 
-    if (n.is_equal(numZERO())) {
-        return exONE();
+    if (n.is_equal(_num0())) {
+        return _ex1();
     }
     
     epvector distrseq;
@@ -667,7 +788,7 @@ ex power::expand_mul(mul const & m, numeric const & n) const
 }
 
 /*
-ex power::expand_commutative_3(ex const & basis, numeric const & exponent,
+ex power::expand_commutative_3(const ex & basis, const numeric & exponent,
                              unsigned options) const
 {
     // obsolete
@@ -675,7 +796,7 @@ ex power::expand_commutative_3(ex const & basis, numeric const & exponent,
     exvector distrseq;
     epvector splitseq;
 
-    add const & addref=static_cast<add const &>(*basis.bp);
+    const add & addref=static_cast<const add &>(*basis.bp);
 
     splitseq=addref.seq;
     splitseq.pop_back();
@@ -695,10 +816,10 @@ ex power::expand_commutative_3(ex const & basis, numeric const & exponent,
 */
 
 /*
-ex power::expand_noncommutative(ex const & basis, numeric const & exponent,
+ex power::expand_noncommutative(const ex & basis, const numeric & exponent,
                                 unsigned options) const
 {
-    ex rest_power=ex(power(basis,exponent.add(numMINUSONE()))).
+    ex rest_power=ex(power(basis,exponent.add(_num_1()))).
                   expand(options | expand_options::internal_do_not_expand_power_operands);
 
     return ex(mul(rest_power,basis),0).
@@ -719,6 +840,15 @@ unsigned power::precedence=60;
 //////////
 
 const power some_power;
-type_info const & typeid_power=typeid(some_power);
+const type_info & typeid_power=typeid(some_power);
 
+// helper function
+
+ex sqrt(const ex & a)
+{
+    return power(a,_ex1_2());
+}
+
+#ifndef NO_NAMESPACE_GINAC
 } // namespace GiNaC
+#endif // ndef NO_NAMESPACE_GINAC
